from .._util.progress import ProgressTracker

from typing import Annotated
from urllib.request import urlretrieve
from zipfile import ZipFile

import typer

def init(
        ctx: typer.Context,
        godot_url: Annotated[str, typer.Option(
            help="The URL to download a ZIP file containing the Windows version of Godot 3.5.3 from"
            # TODO: Linux users don't like being alienated (I should know, I am one)
        )] = "https://github.com/godotengine/godot/releases/download/3.5.3-stable/Godot_v3.5.3-stable_win64.exe.zip",
        gdre_url: Annotated[str, typer.Option(
            help="The URL to download a ZIP file containing the Windows version of GDRE Tools from"
        )] = "https://github.com/bruvzg/gdsdecomp/releases/download/v0.6.2/GDRE_tools-v0.6.2-windows.zip",
):
    logger = ctx.obj["logger"]
    status_tracker = ProgressTracker(logger)

    logger.debug("Ensuring AppData directory exists...")
    mods_directory = ctx.obj["data_directory"] / "mods"
    if not mods_directory.exists():
        mods_directory.mkdir(parents=True)
        logger.info("Created AppData directory")

    logger.debug("Ensuring dependencies directory exists...")
    dependencies_directory = ctx.obj["data_directory"] / "dependencies"
    if not dependencies_directory.exists():
        dependencies_directory.mkdir(parents=True)
        logger.debug("Created dependencies directory")

    # TODO: We might be able to unzip the download while it downloads... but that sounds like WORK
    logger.info("Retrieving Godot from the Internet...")
    godot_zip_path = dependencies_directory / "godot.zip"
    urlretrieve(godot_url, godot_zip_path, status_tracker.print_progress)
    logger.debug("Extracting Godot to the AppData directory...")
    with ZipFile(godot_zip_path, "r") as godot_zip:
        godot_zip.extractall(dependencies_directory)
    logger.debug("Cleaning up Godot ZIP file...")
    godot_zip_path.unlink()

    logger.info("Retrieving GDRE Tools from the Internet...")
    status_tracker.reset()
    gdre_zip_path = dependencies_directory / "gdre_tools.zip"
    urlretrieve(gdre_url, gdre_zip_path, status_tracker.print_progress)
    logger.debug("Extracting GDRE Tools to the AppData directory...")
    with ZipFile(gdre_zip_path, "r") as gdre_zip:
        gdre_zip.extractall(dependencies_directory)
    logger.debug("Cleaning up GDRE Tools ZIP file...")
    gdre_zip_path.unlink()



from .._util.slug import generate_slug

import os
from typing import Annotated

import click
import typer
import yaml

def slug_factory():
    ctx = click.get_current_context()
    name = ctx.params["name"]
    return generate_slug(name)

def create(
        ctx: typer.Context,
        name: Annotated[str, typer.Option("--name", "-n", help="The name of the mod",
                                          prompt="Mod Name", rich_help_panel="Basic Info")],
        creator: Annotated[str, typer.Option("--creator", "-c", help="The creator of the mod",
                                             prompt=True, rich_help_panel="Basic Info")],
        slug: Annotated[str, typer.Option("--slug", "-s",
                                          help="The slug of the mod (terminal-friendly name)",
                                          default_factory=slug_factory, rich_help_panel="Advanced Options")],
        version: Annotated[str, typer.Option("--version", "-v", help="The version of the mod",
                                             rich_help_panel="Detailed Info")] = "0.1.0",
        description: Annotated[str, typer.Option("--description", "-d",
                                                 help="The description of the mod",
                                                 rich_help_panel="Detailed Info")] = "A DnDG Mod",
        export_cards: Annotated[bool, typer.Option(help="Generate files for creating custom cards.",
                                                   rich_help_panel="Advanced Options")] = True,
        export_encounters: Annotated[bool, typer.Option(help="Generate files for creating custom encounters.",
                                                        rich_help_panel="Advanced Options")] = False,
        open_directory: Annotated[bool, typer.Option(help="Open the mod's directory in File Explorer after mod "
                                                          "creation.", rich_help_panel="Advanced Options")] = True,
):
    """Create a blank mod."""
    mod_directory = ctx.obj["data_directory"] / "mods" / slug
    logger = ctx.obj["logger"]
    logger.info(f"Creating '{name}' at {mod_directory}...")

    if mod_directory.exists():
        raise FileExistsError(f"{mod_directory} exists, a new mod cannot be created there.")
    mod_directory.mkdir()
    logger.debug(f"Created mod {name} directory at {mod_directory}")

    with open(mod_directory / "mod.yaml", "w") as f:
        exports = []
        if export_cards:
            exports.append("cards")
        if export_encounters:
            exports.append("encounters")
        data = {
            "name": name,
            "description": description,
            "creator": creator,
            "version": version,
            "exports": exports,
        }
        yaml.safe_dump(data, f, sort_keys=False)
    logger.debug(f"Created mod {name} properties file at {mod_directory / 'mod.yaml'}")

    os.mkdir(mod_directory / "res")
    logger.debug(f"Created mod {name} resources directory at {mod_directory / 'res'}")
    os.mkdir(mod_directory / "src")
    logger.debug(f"Created mod {name} source code directory at {mod_directory / 'src'}")

    with open(mod_directory / "cards.yaml", "w") as f:
        f.write("# Add your cards here!")
    logger.debug(f"Created mod {name} cards file at {mod_directory / 'cards.yaml'}")
    with open(mod_directory / "encounters.yaml", "w") as f:
        f.write("# Add your encounters here!")
    logger.debug(f"Created mod {name} encounters file at {mod_directory / 'encounters.yaml'}")

    logger.info(f"Mod {name} created at {mod_directory}")
    if open_directory:
        os.startfile(mod_directory)



import os
import shutil
from typing import Annotated

import typer
import yaml

def delete(
        ctx: typer.Context,
        slug: Annotated[str, typer.Option("--slug", "-s", help="The slug of the mod to delete",
                                          prompt=True, confirmation_prompt=True)],
):
    """Delete an existing mod."""
    mods_directory = ctx.obj["mods_directory"]
    mod_directory = mods_directory / slug
    logger = ctx.obj["logger"]

    if not os.path.exists(mod_directory):
        raise FileNotFoundError(f"Directory {mod_directory} does not exist!")

    with open(mod_directory / "mod.yaml") as f:
        data = yaml.safe_load(f)
        name = data["name"]

    logger.info(f"Deleting mod {name} at {mod_directory}...")
    shutil.rmtree(mod_directory)
    logger.info(f"Successfully deleted mod {name} at {mod_directory}")



import os
from typing import Annotated

import typer

def open_(
        ctx: typer.Context,
        slug: Annotated[str, typer.Option("--slug", "-s", help="The slug of the mod to open",
                                          prompt=True)],
):
    """Open the directory of a mod in File Explorer."""
    logger = ctx.obj["logger"]
    mod_directory = ctx.obj["data_directory"] / "mods" / slug

    os.startfile(mod_directory)  # TODO: This only works on Windows
    logger.debug(f"Ran os.startfile on {mod_directory}")



class ProgressTracker:
    MULTIPLES = 20

    def __init__(self, logger):
        self.last_step = -1
        self.logger = logger

    def print_progress(self, block_num, block_size, total_size):
        downloaded_size = block_num * block_size
        percentage = 100 * downloaded_size / total_size
        if percentage // self.MULTIPLES > self.last_step:
            self.logger.info(f"{percentage:.0f}% downloaded")
        else:
            self.logger.debug(f"{percentage:.2f}% downloaded")
        self.last_step = percentage // self.MULTIPLES

    def reset(self):
        self.last_step = -1


import re
import unicodedata

def generate_slug(name):
    """Generate a slug from a full name.
        Adapted from https://github.com/django/django/blob/master/django/utils/text.py"""
    slug = unicodedata.normalize('NFKD', name).encode('ascii', 'ignore').decode('ascii')
    slug = re.sub(r'[^\w\s-]', '', slug.lower())
    return re.sub(r'[-\s]+', '-', slug).strip('-_')



from .subcommands import create, delete, open_, init
from ._util.logger import logger_setup, LogLevels

import pathlib
from typing import Annotated

import appdirs
import typer

EPILOG = ("[orange1][bold]DnDGMod by TotallyNotSeth[/] | "
          "[italic]Docs: [link=https://dndgmod.rtfd.io]dndgmod.rtfd.io[/][/]")

app = typer.Typer(
    context_settings={
        "help_option_names": ["--help", "-h"]
    },
    rich_markup_mode="rich",
    no_args_is_help=True
)
app.command(epilog=EPILOG)(init.init)
app.command(epilog=EPILOG)(create.create)
app.command("open", epilog=EPILOG)(open_.open_)
app.command(epilog=EPILOG)(delete.delete)

@app.callback()
def dndgmod_callback(
        ctx: typer.Context,
        log_level: Annotated[LogLevels, typer.Option("--log-level", "-l")] = "info",
):
    """
    [underline]DnDGMod: The Dungeons & Degenerate Gamblers Modloader[/]

    Created by TotallyNotSeth in California with :heart:
    Check out our documentation at [link=https://dndgmod.rtfd.io]dndgmod.rtfd.io[/] for help getting started!
    """
    data_directory = pathlib.Path(appdirs.user_data_dir("DnDGMod", "TotallyNotSeth")).resolve()
    mods_directory = data_directory / "mods"
    if not (mods_directory.exists() or ctx.invoked_subcommand == "init"):
        raise FileNotFoundError(f"Directory {mods_directory} doesn't exist! Try running `dndgmod init`")

    ctx.ensure_object(dict)
    ctx.obj["data_directory"] = data_directory
    logger_setup(ctx, log_level)


if __name__ == "__main__":
    app()



import enum
import logging

LOG_LEVEL_MAPPING = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
}

class LogLevels(enum.Enum):
    debug = "debug"
    info = "info"
    warning = "warning"
    error = "error"


def logger_setup(ctx, log_level):
    logger = logging.getLogger("dndgmod")
    logger.setLevel(LOG_LEVEL_MAPPING[log_level.value])
    logger.addHandler(logging.StreamHandler())
    ctx.obj["logger"] = logger
